@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2000-2007 SWsoft. All rights reserved.
+ *  Copyright (C) 2000-2008, Parallels, Inc. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -493,6 +493,33 @@ void convert_quota_stat( void *dest, int dest_ver, void *src, int src_ver)
 	}
 }
 
+/*
+ * Show hints and error message in case of quota on failed.
+ */
+static void error_quotaon(struct qf_data *qd, int err, char *buf)
+{
+	/* show error message */
+	error(0, err, "Quota on syscall for id %d", quota_id);
+
+	/* try to find the reason */
+	if (err == EBUSY) {
+		error(0, 0, "\tPossible reasons:");
+		error(0, 0, "\t- Container's root is already mounted");
+		error(0, 0, "\t- there are opened files inside Container's private"
+				" area");
+		error(0, 0, "\t- your current working directory is inside Container's");
+		error(0, 0, "\t  private area");
+		if (buf) {
+			error(0, 0, "\tCurrently used file(s):\n%s", buf);
+			if (strlen(buf) > getpagesize() - 2)
+				error(0, 0, "\t... more files found");
+		}
+	}
+
+	/* exit with error status */
+	error(EC_VZCALL, 0, NULL);
+}
+
 /* turn quota on
  * return 0 if success,
  * <0 if quota is on */
@@ -504,6 +531,7 @@ int quota_syscall_on(struct qf_data *qd)
 	struct vz_quota_stat *stat;
 	struct ugid_quota *ugid_stat;
 	char *path;
+	char *buf;
 
 	ASSERT(qd);
 	stat = &qd->stat;
@@ -625,17 +653,27 @@ int quota_syscall_on(struct qf_data *qd)
 		ugid_stat->info.buf_size = ugid_stat->dquot_size;
 	}
 
+	buf = NULL;
+
 	/* turn quota on */
 	for (retry = 0; ; retry++) {
-		rc = vzquotactl_syscall(VZ_DQ_ON, quota_id, NULL, path);
+		 /*
+		  * If this is the last chance to enable quota, allocate memory buffer
+		  * and ask kernel to find currently used files.
+		  */
+		if (retry == MAX_RETRY) {
+			buf = (char *)malloc(getpagesize());
+			*buf = '\0';
+		}
+		rc = vzquotactl_syscall(VZ_DQ_ON, quota_id,
+					(struct vz_quota_stat *)buf, path);
 		if (rc >= 0) break;
 		if (errno != EBUSY || retry >= MAX_RETRY) {
 			int save_errno;
 
 			save_errno = errno;
 			vzquotactl_syscall(VZ_DQ_DESTROY, quota_id, NULL, NULL);
-			error(EC_VZCALL, save_errno, "Quota on syscall for %d", 
-					quota_id);
+			error_quotaon(qd, save_errno, buf);
 		}
 		usleep(sleeptime * 1000);
 		sleeptime = sleeptime * 2;
@@ -650,6 +688,40 @@ int quota_syscall_on(struct qf_data *qd)
 	return 0;
 }
 
+/*
+ * Show hints and error message in case of quota off failed.
+ */
+static void error_quotaoff(struct qf_data *qd, int err, char *buf)
+{
+	/* show error message */
+	error(0, err, "Quota off syscall for id %d", quota_id);
+
+	/* try to find the reason */
+	if (err == EBUSY && (option & FL_FORCE))
+		/* Hmm, shouldn't be */
+		error(0, 0, "Forced quota off failed. Device is busy\n");
+	else if (err == EBUSY) {
+		error(0, 0, "\tPossible reasons:");
+		error(0, 0, "\t- Container's root is not unmounted");
+		error(0, 0, "\t- there are opened files inside container's root/private"
+				" area");
+		error(0, 0, "\t- your current working directory is inside a container directory");
+		error(0, 0, "\t  root/private area");
+		if (buf) {
+			error(0, 0, "\tCurrently used file(s):\n%s", buf);
+			if (strlen(buf) > getpagesize() - 2)
+				error(0, 0, "\t... more files found");
+		}
+	} else if (err == EINVAL && (option & FL_FORCE)) {
+		error(0, 0, "\tPossible reasons:");
+		error(0, 0, "\t- Kernel doesn't support forced quota off");
+	}
+
+	/* exit with error status */
+	error(EC_VZCALL, 0, NULL);
+
+}
+
 /* turn quota off
  * return 0 if success,
  * OENT if quota is off */
@@ -661,14 +733,27 @@ int quota_syscall_off(struct qf_data *qd)
 	int broken = 0;
 	struct vz_quota_stat *stat;
 	struct ugid_quota *ugid_stat;
+	int cmd;
+	char *buf;
 
 	ASSERT(qd);
 	stat = &qd->stat;
 	ugid_stat = &qd->ugid_stat;
-	
+
+	cmd = VZ_DQ_OFF;
+	buf = NULL;
+
 	/* turn quota off */
 	for (retry = 0; ; retry++) {
-		rc = vzquotactl_syscall(VZ_DQ_OFF, quota_id, NULL, NULL);
+		/*
+		 * If this is the last chance to disable quota, allocate memory buffer
+		 * and ask kernel to find currently used files.
+		 */
+		if (retry == MAX_RETRY && debug_level >= LOG_INFO) {
+			buf = (char *)malloc(getpagesize());
+			*buf = '\0';
+		}
+		rc = vzquotactl_syscall(cmd, quota_id, NULL, buf);
 		if (rc >= 0) break;
 		if (errno == EALREADY) { /* quota is created and off */
 			broken = 1;
@@ -676,9 +761,20 @@ int quota_syscall_off(struct qf_data *qd)
 		}
 		if (errno == ENOENT) return rc;		/* quota is not created */
 		if (errno != EBUSY || retry >= MAX_RETRY)
-			error(EC_VZCALL, errno, "Quota off syscall for id %d", quota_id);
+			error_quotaoff(qd, errno, buf);
 		usleep(sleeptime * 1000);
 		sleeptime = sleeptime * 2;
+
+		/*
+		 * When 'force' option is set, try to use VZ_DQ_OFF first,
+		 * and in case of failure use VZ_DQ_OFF_FORCED.
+		 */
+		if (errno == EBUSY && (option & FL_FORCE) &&
+				(retry >= (MAX_RETRY - 1))) {
+			cmd = VZ_DQ_OFF_FORCED;
+			/* Mark as dirty now */
+			qd->head.flags |= QUOTA_DIRTY;
+		}
 	}
 	
 	/* get VE stat */
